/*
 * @(#)PhraseEdit.h 3.00 19 May 1999
 *
 * Copyright (c) 2000 Pete Goodliffe (pete.goodliffe@pace.co.uk)
 *
 * This file is part of TSE3 - the Trax Sequencer Engine version 3.00.
 *
 * This library is modifiable/redistributable under the terms of the GNU
 * General Public License.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef TSE3_PHRASEEDIT_H
#define TSE3_PHRASEEDIT_H

#include "tse3/MidiData.h"

#include <string>

namespace TSE3
{
    class Phrase;

    /**
     * @ref PhraseEdit listener interface.
     *
     * @author  Pete Goodliffe
     * @version 3.00
     * @see     Listener
     */
    class PhraseEditListener : public MidiDataListener
    {
        public:
            typedef PhraseEdit notifier_type;

            virtual void PhraseEdit_Reset(PhraseEdit *)                   {}
            virtual void PhraseEdit_Tidied(PhraseEdit *)                  {}
            virtual void PhraseEdit_Inserted(PhraseEdit *, int /*index*/) {}
            virtual void PhraseEdit_Erased(PhraseEdit *, int /*index*/)   {}
    };

    /**
     * This class provides an editable @ref MidiData area. It is the mechanism
     * used to generate @ref Phrases.
     *
     * The state of the PhraseEdit area can be one of:
     *
     * @li Tidy: Sustain pedal useage has been converted into sustained notes
     * and MidiCommand_NoteOffs are coalescened into their appropriate
     * MidiEvents.
     *
     * @li Untidy: MidiCommand_NoteOffs may not be matched into MidiEvents
     * and all hell could generally break loose if you try to play this MidiData
     * in it's current state.
     *
     * It is up to the user to know if a PhraseEdit area may be untidy and to
     * deal with it appropriately (i.e. make sure it is never played).
     *
     * @short   An editable Phrase
     * @author  Pete Goodliffe
     * @version 3.00
     * @see     MidiData
     * @see     Phrase
     */
    class PhraseEdit : public MidiData,
                       public Notifier<PhraseEditListener>
    {
        public:

            /**
             * The deault maximum no events in ~ref Phrase editor. These days
             * the editor automatically grows so this is fairly redundant.
             */
            static const int defaultSize = 1024*10;

            /**
             * The time allowance given for notes to be before start time when
             * tidying.
             *
             * @see tidy
             */
            static const int tollerance = Clock::PPQN/2;

            /**
             * Create a PhraseEdit with the given number of events
             * space reserved.
             *
             * It is not really necessary to specify this size unless you know
             * at least how many events will be inserted and reserving them
             * will result in a worthwhile performance increase.
             *
             * @param noEvents Ignorable PhraseEdit size
             */
            explicit PhraseEdit(int noEvents = defaultSize);

            virtual ~PhraseEdit();

            /**
             * Resets the PhraseEdit area to be a copy of the given MidiData.
             * If the @p source pointer is zero, the PhraseEdit area will be
             * blank.
             *
             * @param source @ref MidiData to copy
             */
            void reset(const MidiData *source = 0);

            /**
             * Tidies the @ref MidiData (usually used after recording has
             * stopped) This does a number of things, including adding any
             * missing MidiCommand_NoteOffs and the conversion of sustain pedal
             * usage into elongated notes. It removes all events before
             * time zero (with a small tollerance)
             *
             * The data in the PhraseEdit at this point could be in any
             * state. MidiCommand_NoteOffs may be collated into the correct
             * @ref MidiEvent, or they may be in their own event (which they
             * will be when just recorded). Events may be in time order, or
             * in some completely random order.
             *
             * It also collates all MidiCommand_NoteOffs into the
             * appropriate MidiCommand_NoteOns so that the PhraseEdit
             * area will contain valid data that can be streamed out through
             * a @ref Playable interface.
             *
             * @param stopTime The time of the end of this MidiData (e.g. the
             *                 time recording stopped). If no stopTime is
             *                 specified then it is inferred from the last
             *                 event in the MidiData.
             * @see   tollerance
             */
            void tidy(Clock stopTime = -1);

            /**
             * Creates a @ref Phrase from this @ref MidiData. Make sure that
             * you have done a tidy() first.
             *
             * The @ref Phrase object is created with new; it is your
             * responsibility to ensure it is deleted afterwards.
             *
             * @return New @ref Phrase object spawned from the data in this
             *         PhraseEdit - you must delete this
             * @see    tidy
             * @see    Phrase
             */
            Phrase *createPhrase() const;

            /**
             * Returns the nth @ref MidiEvent in this @ref MidiData object.
             *
             * Note that, unlike the base @ref MidiData class, this object
             * is modifiable.
             *
             * The value returned for an index that is out of range is
             * undefined. The @ref size method describes the valid
             * values.
             *
             * @param  n Index
             * @return @ref MidiEvent at index n
             */
            MidiEvent &operator[](size_t n) { return data[n]; }

            /**
             * Insert a @ref MidiEvent into the @ref MidiData in time order.
             *
             * @sect Important note
             *
             * If this is a @ref MidiCommand_NoteOn then you really want the
             * @ref MidiEvent to also hold the ballancing
             * @ref MidiCommand_NoteOff in its 'second half'.
             *
             * If it does then you can be assured that adding the
             * @ref MidiEvent to a 'tidied' PhraseEdit object will leave the
             * object still 'tidy'. If you only add a @ref MidiCommand_NoteOn/
             * @ref MidiCommand_NoteOff without the ballancing MIDI command
             * then the PhraseEdit is left 'untidy' and any use of it to create
             * a @ref Phrase and/or have it's data streamed out of a
             * @ref Playable interface is invalid.
             *
             * This restriction is important in the use of the TSE3 library,
             * if you don't use 'tidy' @ref MidiData then playback will produce
             * 'undefined' (very probably wrong) output.
             *
             * @param MidiEvent to insert.
             * @see   tidy
             */
            void insert(MidiEvent event);

            /**
             * Remove a @ref MidiEvent from the @ref MidiData.
             *
             * @param n Index of MidiEvent to remove.
             */
            void erase(size_t n);

            /**
             * Remove a @ref MidiEvent from the @ref MidiData. If there isn't a
             * @ref MidiEvent * matching this one then no change occurs.
             *
             * @param event MidiEvent to remove.
             */
            void erase(MidiEvent event);

        private:

            PhraseEdit &operator=(const PhraseEdit &);
            PhraseEdit(const PhraseEdit &);

            /**
             * hint is a cached index that refers to the last position of
             * intertion. Since most inserts are performed in time order this
             * cached information can increase performance drastically.
             *
             * It has been shown, for example, to reduce MIDI file import
             * by at least a third.
             *
             * This used to be an iterator, however the insertion of the
             * iterator made it invalid which could lead to some freaky
             * crashes now and again.
             */
            size_t hint;

    };
}

#endif
