/*
 * @(#)Phrase.h 3.00 21 November 2000
 *
 * Copyright (c) 2000 Pete Goodliffe (pete.goodliffe@pace.co.uk)
 *
 * This file is part of TSE3 - the Trax Sequencer Engine version 3.00.
 *
 * This library is modifiable/redistributable under the terms of the GNU
 * General Public License.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef TSE3_CMD_PHRASE_H
#define TSE3_CMD_PHRASE_H

#include "tse3/cmd/Command.h"
#include "tse3/Phrase.h"
#include "tse3/DisplayParams.h"

#include <vector>

namespace TSE3
{
    class Phrase;
    class Song;
    class Part;

    namespace Cmd
    {
        /**
         * Command to set a @ref TSE3::Phrase's info (the title and its
         * @ref TSE3::DisplayParams setup).
         *
         * If the @ref TSE3::Phrase is parented in a @ref TSE3::PhraseList
         * and you change the title to an invalid name, then the exception
         * generated by @ref TSE3::PhraseList::insert() will be propragated back
         * to you.
         *
         * @short   Set @ref TSE3::Phrase info Command
         * @author  Pete Goodliffe
         * @version 1.00
         * @see     Command
         */
        class Phrase_SetInfo : public Command
        {
            public:

                /**
                 * To create this command specify the @ref TSE3::Track object to
                 * alter and the new information.
                 */
                Phrase_SetInfo(TSE3::Phrase              *phrase,
                               const std::string         &title,
                               const TSE3::DisplayParams &dp);

            protected:

                /**
                 * @reimplemented
                 */
                virtual void executeImpl();

                /**
                 * @reimplemented
                 */
                virtual void undoImpl();

            private:

                TSE3::Phrase        *phrase;
                std::string          newTitle, oldTitle;
                TSE3::DisplayParams  dp;
        };

        /**
         * Command to insert a @ref TSE3::Phrase into a @ref TSE3::PhraseList.
         *
         * This object may propagate the @ref TSE3::PhraseList exception if
         * the @ref Phrase name already exists.
         *
         * @short   Insert @ref TSE3::Phrase Command
         * @author  Pete Goodliffe
         * @version 3.00
         * @see     Command
         */
        class Phrase_Insert : public Command
        {
            public:

                /**
                 * Creates a new Phrase_Insert command.
                 *
                 * @param phrase     @ref TSE3::Phrase to insert
                 * @param phraseList @ref TSE3::PhraseList to insert it in
                 */
                Phrase_Insert(TSE3::Phrase     *phrase,
                              TSE3::PhraseList *phraseList);
                virtual ~Phrase_Insert();

            protected:

                /**
                 * @reimplemented
                 */
                virtual void executeImpl();

                /**
                 * @reimplemented
                 */
                virtual void undoImpl();

            private:

                TSE3::Phrase        *phrase;
                TSE3::PhraseList    *phraseList;
        };

        /**
         * Command to remove a @ref TSE3::Phrase from a @ref TSE3::PhraseList.
         *
         * @short   Remove @ref TSE3::Phrase Command
         * @author  Pete Goodliffe
         * @version 3.00
         * @see     Command
         */
        class Phrase_Remove : public Command
        {
            public:

                /**
                 * Creates a new Phrase_Remove command.
                 *
                 * The @ref TSE3::PhraseList to remove from is read from the
                 * @ref TSE3::Phrase.
                 *
                 * @param phrase @ref TSE3::Phrase to remove
                 * @param song   @ref TSE3::Song to remove from, if you
                 *               specify this then every use of the
                 *               @ref TSE3::Phrase in a @ref TSE3::Part in the
                 *               @ref TSE3::Song will be removed.
                 */
                Phrase_Remove(TSE3::Phrase *phrase,
                              TSE3::Song   *song = 0);
                virtual ~Phrase_Remove();

            protected:

                /**
                 * @reimplemented
                 */
                virtual void executeImpl();

                /**
                 * @reimplemented
                 */
                virtual void undoImpl();

            private:

                TSE3::Phrase              *phrase;
                TSE3::Song                *song;
                std::vector<TSE3::Part *>  parts;
                bool                       vector_done;
        };

        /**
         * Command to replace a @ref TSE3::Phrase in a @ref TSE3::Song with
         * another @ref TSE3::Phrase.
         *
         * If both Phrases have a parent @ref TSE3::PhraseList (which should be
         * the same @ref TSE3::PhraseList, from the same @ref TSE3::Song) then
         * every @ref TSE3::Part that uses the @p oldPhrase will be made to use
         * the @p newPhrase.
         *
         * However, if the @p newPhrase has no parent @ref TSE3::PhraseList
         * then the replacement will be slightly different. The @p oldPhrase
         * is removed from the @ref TSE3::PhraseList completely. The @p
         * newPhrase is inserted in its place. If the @p newPhrase has not
         * yet been given a title, then it will be give the same title as the
         * @p oldPhrase. The replacement operation will be performed, and the
         * the @p newPhrase will be inserted into the @ref TSE3::PhraseList.
         *
         * @short   Replace @ref TSE3::Phrase Command
         * @author  Pete Goodliffe
         * @version 3.00
         * @see     Command
         */
        class Phrase_Replace : public Command
        {
            public:

                /**
                 * Creates a new Phrase_Replace command.
                 *
                 * @param oldPhrase @ref TSE3::Phrase to remove
                 * @param newPhrase @ref TSE3::Phrase to replace with
                 * @param song      @ref TSE3::Song to act on
                 */
                Phrase_Replace(TSE3::Phrase *oldPhrase,
                               TSE3::Phrase *newPhrase,
                               TSE3::Song   *song);
                ~Phrase_Replace();

            protected:

                /**
                 * @reimplemented
                 */
                virtual void executeImpl();

                /**
                 * @reimplemented
                 */
                virtual void undoImpl();

            private:

                TSE3::Phrase              *newPhrase;
                TSE3::Phrase              *oldPhrase;
                bool                       newPhraseParented;
                TSE3::Song                *song;
        };
    }
}

#endif
