/*
 * @(#)OSS.h 3.00 20 July 1999
 *
 * Copyright (c) 2000 Pete Goodliffe (pete.goodliffe@pace.co.uk)
 *
 * This file is part of TSE3 - the Trax Sequencer Engine version 3.00.
 *
 * This library is modifiable/redistributable under the terms of the GNU
 * General Public License.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef TSE3_PLT_OSS_H
#define TSE3_PLT_OSS_H

#include "tse3/MidiScheduler.h"

#include <list>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef TSE3_WITH_OSS
#include <sys/soundcard.h>
#else
    struct synth_info;
    struct midi_info;
#endif

namespace TSE3
{
    namespace Plt
    {
        /**
         * An class that provides support for dynamic voice id allocation for
         * soundcard MIDI output emulation.
         *
         * MIDI devices will accept MidiCommand_NoteOns as they come in. If
         * they get too many then they will have a policy to deal with it:
         * perhaps not play the additional notes, perhaps switch off the oldest
         * note, perhaps something more elaborate. Soundcards, on the other
         * hand, generally have a fixed number of voice generators that one
         * must address individually. In this case we need a list of which
         * voices are not yet in use to play a new note along side others. We
         * also need a list of which voice is assigned to which note (and with
         * which patch) so that we can switch off the correct voice. That's the
         * functionality provided by this class.
         *
         * It is used by versions of the @ref OSSMidiScheduler_SynthDevice
         * class to manage which voices are being played at any time.
         *
         * This is an internal class - you don't need to directly manipulate it
         * unless you specifically require it's functionality.
         *
         * @short   Soundcard voice id allocator
         * @author  Pete Goodliffe
         * @version 0.00
         * @internal
         */
        class VoiceManager
        {
            public:

                /**
                 * Create a VoiceManager that will handle noVoices voices.
                 */
                VoiceManager(int noVoices);
                ~VoiceManager();

                /**
                 * Allocate a voice from those available for the
                 * given soundcard. Returns an ID handle for the voice.
                 *
                 * It will first attempt to allocate a currently unused voice.
                 * If there are no unused voices then it will reallocate the
                 * last used voice.
                 *
                 * This method will not perform any OSS system calls to set the
                 * voice up.
                 */
                int allocate(int channel, int note);

                /**
                 * Deallocate a previously allocated voice.
                 *
                 * This method will not perform any OSS system calls to shut
                 * a voice down - it must have been stopped before deallocate
                 * is called.
                 */
                void deallocate(int id);

                /**
                 * Perform a search throught the currently allocated voices
                 * for one matching the given channel.
                 *
                 * Call first time with pos = -1. If returns -1 then there
                 * are no voices on the given channel. Call subsequently
                 * with pos set to the return ID to get the next match.
                 */
                int search(int channel, int pos);

                /**
                 * Similar to search above, but includes the note as a
                 * search criteria.
                 */
                int search(int channel, int note, int pos);

            private:

                int noVoices;
                struct Voice
                {
                    const int  id;
                    int        channel;
                    int        note;
                    bool       used;
                    Voice(int id) : id(id), used(false) {}
                };
                Voice             **voices;       // array of all Voice structs
                std::list<Voice*>   usedVoices;   // used voices pushed on back
                std::list<Voice*>   unusedVoices; // unused voices stored here
        };

        /**
         * A base class for classes that handle different types of OSS device.
         *
         * The class uses a number of member variables of the
         * @ref OSSMidiScheduler object - these are the OSS interface data
         * memebers that the OSS macros write to. They either have to be shared
         * in this manner, or the functionality for each type of device must be
         * rolled into the @ref OSSMidiScheduler class (which is the least
         * desirable of the two).
         *
         * This is an internal class - you don't need to directly manipulate it
         * unless you specifically require it's functionality.
         *
         * @short   Soundcard playback handler
         * @author  Pete Goodliffe
         * @version 0.00
         * @internal
         */
        class OSSMidiScheduler_SynthDevice
        {
            public:

                /**
                 * Constructor. Pass in the OSS interface data and the
                 * device number that the class is to provice support for.
                 */
                OSSMidiScheduler_SynthDevice(int             deviceno,
                                             synth_info     &synthinfo,
                                             int             seqfd,
                                             unsigned char *&_seqbuf,
                                             int            &_seqbuflen,
                                             int            &_seqbufptr);
                virtual ~OSSMidiScheduler_SynthDevice() = 0;

                virtual void noteOff(int ch, int note, int vel) = 0;
                virtual void noteOn(int ch, int note, int vel) = 0;
                virtual void keyPressure(int ch, int note, int vel) = 0;
                virtual void controlChange(int ch, int ctrl, int val) = 0;
                virtual void programChange(int ch, int prog) = 0;
                virtual void channelPressure(int ch, int vel) = 0;
                virtual void pitchBend(int ch, int lsb, int msb) = 0;

            protected:

                int deviceno; // OSS device the class handles

                // These data members are created by the OSSMidiScheduler class
                // and we use them here. We do not delete them in the
                // constructor.
                int             seqfd;
                synth_info     &synthinfo;
                unsigned char *&_seqbuf;
                int            &_seqbuflen;
                int            &_seqbufptr;
                void            seqbuf_dump();
                void            seqbuf_clean();

                // Cached information about each channel's state
                unsigned char _programChange[16];
                unsigned char _pitchBendLSB[16];
                unsigned char _pitchBendMSB[16];
                unsigned char _channelPressure[16];
        };


        /**
         * OSS NULL device handler - it does nothing.
         *
         * This is an internal class, used by the @ref OSSMidiScheduler class.
         *
         * @short   Unknown soundcard playback handler
         * @author  Pete Goodliffe
         * @version 0.00
         * @internal
         */
        class OSSMidiScheduler_NULLDevice : public OSSMidiScheduler_SynthDevice
        {
            public:

                OSSMidiScheduler_NULLDevice(int            deviceno,
                                           synth_info     &synthinfo,
                                           int             seqfd,
                                           unsigned char *&_seqbuf,
                                           int            &_seqbuflen,
                                           int            &_seqbufptr);
                virtual ~OSSMidiScheduler_NULLDevice();

                /**
                 * @reimplemented
                 */
                virtual void noteOff(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void noteOn(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void keyPressure(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void controlChange(int ch, int ctrl, int val);

                /**
                 * @reimplemented
                 */
                virtual void programChange(int ch, int prog);

                /**
                 * @reimplemented
                 */
                virtual void channelPressure(int ch, int vel);

                /**
                 * @reimplemented
                 */
                virtual void pitchBend(int ch, int lsb, int msb);
        };

        /**
         * OSS FM synth device handler.
         *
         * This is an internal class, used by the @ref OSSMidiScheduler class.
         * However, you <b>will</b> need to call @ref setPatchesDirectory()
         * before you create an OSSMidiScheduler.
         *
         * @short   FM soundcard playback handler
         * @author  Pete Goodliffe
         * @version 0.00
         * @internal
         */
        class OSSMidiScheduler_FMDevice : public OSSMidiScheduler_SynthDevice
        {
            public:

                OSSMidiScheduler_FMDevice(int             deviceno,
                                          synth_info     &synthinfo,
                                          int             seqfd,
                                          unsigned char *&_seqbuf,
                                          int            &_seqbuflen,
                                          int            &_seqbufptr);
                virtual ~OSSMidiScheduler_FMDevice();

                /**
                 * @reimplemented
                 */
                virtual void noteOff(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void noteOn(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void keyPressure(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void controlChange(int ch, int ctrl, int val);

                /**
                 * @reimplemented
                 */
                virtual void programChange(int ch, int prog);

                /**
                 * @reimplemented
                 */
                virtual void channelPressure(int ch, int vel);

                /**
                 * @reimplemented
                 */
                virtual void pitchBend(int ch, int lsb, int msb);

                /**
                 * This returns in which directory the device will look for
                 * patch files. This must be set before the device is created.
                 *
                 * The patches directory path can contain multiple paths,
                 * separated by colons.
                 *
                 * @see setPatchesDirectory
                 */
                static std::string &patchesDirectory()
                {
                    return _patchesDirectory;
                }

                /**
                 * Call this before creating the @ref OSSMidiScheduler object
                 * to set up where to look for FM patches files.
                 *
                 * @see patchesDirectory
                 */
                static void setPatchesDirectory(const std::string &dir);

            private:

                VoiceManager voiceman;

                bool patchLoaded[256];      // 0-127 patches, 128-255 drums.
                void loadPatches();         // Load the patches into the device.
                int  getPatch(int patchNo); // returns patchNo if that patchNo
                                            // is loaded, else another patch no
                                            // or 256 if couldn't find any.
                static std::string _patchesDirectory;
                int opl;       // for FM devices whether to do opl 3 (== 3)
        };

        /**
         * OSS AWE synth device handler.
         *
         * This is an internal class, used by the @ref OSSMidiScheduler class.
         *
         * @short   AWE soundcard playback handler
         * @author  Pete Goodliffe
         * @version 0.00
         * @internal
         */
        class OSSMidiScheduler_AWEDevice : public OSSMidiScheduler_SynthDevice
        {
            public:

                OSSMidiScheduler_AWEDevice(int             deviceno,
                                           synth_info     &synthinfo,
                                           int             seqfd,
                                           unsigned char *&_seqbuf,
                                           int            &_seqbuflen,
                                           int            &_seqbufptr);
                virtual ~OSSMidiScheduler_AWEDevice();

                /**
                 * @reimplemented
                 */
                virtual void noteOff(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void noteOn(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void keyPressure(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void controlChange(int ch, int ctrl, int val);

                /**
                 * @reimplemented
                 */
                virtual void programChange(int ch, int prog);

                /**
                 * @reimplemented
                 */
                virtual void channelPressure(int ch, int vel);

                /**
                 * @reimplemented
                 */
                virtual void pitchBend(int ch, int lsb, int msb);
        };

        /**
         * OSS GUS synth device handler.
         *
         * This is an internal class, used by the @ref OSSMidiScheduler class.
         * However, you <b>will</b> need to call @ref setPatchesDirectory()
         * before you create an OSSMidiScheduler.
         *
         * @short   GUS soundcard playback handler
         * @author  Pete Goodliffe
         * @version 0.00
         * @internal
         */
        class OSSMidiScheduler_GUSDevice : public OSSMidiScheduler_SynthDevice
        {
            public:

                OSSMidiScheduler_GUSDevice(int             deviceno,
                                           synth_info     &synthinfo,
                                           int             seqfd,
                                           unsigned char *&_seqbuf,
                                           int            &_seqbuflen,
                                           int            &_seqbufptr);
                virtual ~OSSMidiScheduler_GUSDevice();

                /**
                 * @reimplemented
                 */
                virtual void noteOff(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void noteOn(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void keyPressure(int ch, int note, int vel);

                /**
                 * @reimplemented
                 */
                virtual void controlChange(int ch, int ctrl, int val);

                /**
                 * @reimplemented
                 */
                virtual void programChange(int ch, int prog);

                /**
                 * @reimplemented
                 */
                virtual void channelPressure(int ch, int vel);

                /**
                 * @reimplemented
                 */
                virtual void pitchBend(int ch, int lsb, int msb);

                /**
                 * This returns in which directory the device will look for
                 * patch files. This must be set before the device is created.
                 * (In fact, it must be set before each voice is loaded - they
                 * are loaded on demand. However, it is safest to set this
                 * before creating the device.)
                 *
                 * The patches directory path can contain multiple paths,
                 * separated by colons.
                 *
                 * @see setPatchesDirectory
                 */
                static std::string &patchesDirectory()
                {
                    return _patchesDirectory;
                }

                /**
                 * Call this before creating the @ref OSSMidiScheduler object
                 * to set up where to look for GUS patches files.
                 *
                 * @see patchesDirectory
                 */
                static void setPatchesDirectory(const std::string &dir);

            private:

                VoiceManager voiceman;

                size_t  nobits;                 // patch resolution; 8 or 16
                int     totalMemory;            // total memory in soundcard
                int     freeMemory;             // free memory in soundcard
                bool    patchLoaded[256];       // 0-127 patches, 128-255 drums.
                bool    patchLoadedFailed[256]; // whether a load attempy failed
                int     getPatch(int patchNo);  // returns patchNo if that
                                                // patchNo is loaded, else
                                                // another patch no or 256 if
                                                // couldn't find any.

                static std::string _patchesDirectory;

                /**
                 * Returns the filename for the given program change number.
                 * Calls @ref loadPatch() if the patch hasn't been loaded yet.
                 */
                static char *patchName(int pgm);

                /**
                 * Load a patch. 0..127 is a GM program, 128-255 is a drum
                 * patch. Sets patchLoaded and patchLoadedFailed appropriately.
                 *
                 * @return Boolean indicating sucess.
                 */
                bool loadPatch(int patch);
        };

        /**
         * This is the Open Sound System @ref MidiScheduler implementation.
         * It requires OSS version 3.60.
         *
         * Since it opens /dev/sequencer (the OSS sequencer interface) there
         * can only ever be one of these objects instantiated at any given time.
         * Subsequent instanitations will throw an exception.
         *
         * Before you create an object of this type you may want to call the
         * following static member functions to set the location of patches
         * files:
         *
         * @li @ref OSSMidiScheduler_FMDevice::setPatchesDirectory
         * @li @ref OSSMidiScheduler_GUSDevice::setPatchesDirectory
         *
         * @short   OSS MidiScheduler implemtation
         * @author  Pete Goodliffe
         * @version 0.00
         * @see     MidiScheduler
         * @see     OSSMidiSchedulerFactory
         */
        class OSSMidiScheduler : public MidiScheduler
        {
            public:

                /**
                 * @throws TSE3::MidiSchedulerError
                 */
                OSSMidiScheduler();
                virtual ~OSSMidiScheduler();

                /**
                 * @reimplemented
                 */
                virtual const char *implementationName();

                /**
                 * @reimplemented
                 */
                virtual size_t ports() const;

                /**
                 * @reimplemented
                 */
                virtual const char *portName(size_t port) const;

                /**
                 * @reimplemented
                 */
                virtual const char *portType(size_t port) const;

                /**
                 * @reimplemented
                 */
                virtual bool portReadable(size_t port) const;

                /**
                 * @reimplemented
                 */
                virtual bool portWriteable(size_t port) const;

                /**
                 * @reimplemented
                 */
                virtual void tx(MidiCommand mc);

                /**
                 * @reimplemented
                 */
                virtual void start(Clock startTime);

                /**
                 * @reimplemented
                 */
                virtual void stop(Clock stopTime);

                /**
                 * @reimplemented
                 */
                virtual void moveTo(Clock moveTime, Clock newTime);

                /**
                 * @reimplemented
                 */
                virtual Clock clock();

                /**
                 * @reimplemented
                 */
                virtual int msecs();

                /**
                 * @reimplemented
                 */
                virtual void setTempo(int newTempo, Clock changeTime);

                /**
                 * @reimplemented
                 */
                virtual bool eventWaiting();

                /**
                 * @reimplemented
                 */
                virtual MidiEvent rx();

                /**
                 * @reimplemented
                 */
                virtual void tx(MidiEvent event);

                /**
                 * @reimplemented
                 */
                virtual void txSysEx(const unsigned char *data, size_t size);

            private:

                OSSMidiScheduler(const OSSMidiScheduler &);
                OSSMidiScheduler &operator=(const OSSMidiScheduler &);

                /**
                 * Does the tx. If outOfBand is false, sends data to the
                 * end of the playback queue. If outOfBand is true, it
                 * sends data immediately.
                 */
                void tx(const MidiCommand mc, bool outOfBand);

                unsigned int  nosynths;    // no of internal synth devices
                unsigned int  nomidis;     // no of MIDI ports
                unsigned int  nodevices;   // nosynths + nomidis
                int           rate;        // sequencer timer rate (in ms)
                int           rateDivisor; // value to divide timer by to get ms
                synth_info   *synthinfo;   // synth_info for each synth device
                midi_info    *midiinfo;    // midi_info for each MIDI device

                // In C we write SEQ_DEFINEBUF(1024); to set up some OSS magic
                // which creates a buffer that the OSS macros write data into.
                // However, this won't work in a class declaration. Sigh.
                // So this is how we do it.
                int            seqfd;          // file descfor /dev/sequencer
                unsigned char *_seqbuf;        // buffer for /dev/sequencer data
                int            _seqbuflen;     // size of _seqbuf
                int            _seqbufptr;     // pointer to end of pending data
                void           seqbuf_dump();  // wrts _seqbuf to /dev/sequencer
                void           seqbuf_clean(); // flushes the buffer

                // For a given port (TSE3 speak, same as OSS device number) is
                // it a synth or MIDI device?
                bool isSynth(unsigned int port) const
                {
                    return port < nosynths;
                }
                bool isMidi(unsigned int port) const
                {
                    return port >= nosynths;
                }

                // MIDI devices are easy to send data too, but the synths need
                // some special massaging. The classes held in this table do
                // just that.
                OSSMidiScheduler_SynthDevice **devices;

                // Output malarky
                Clock lastTxTime; // Time last MidiCommand was scheduled

                // Input malarky
                bool        input;   // true if there is an input command
                MidiCommand command; // if input is true, this holds the input
                Clock       time;    // time of the input MidiCommand

                /**
                 * Performs input scan of the /dev/sequencer device and
                 * interprets what it reads. If there is any input, then
                 * the input variable is set to true. When you've read the
                 * input, then set it to false.
                 */
                void readInput();
        };

        /**
         * This is the Open Sound System version of the
         * @ref MidiSchedulerFactory. It makes appropriate decisions to deliver
         * you a @ref MidiScheduler that will run under OSS version 3.60.
         *
         * @short   OSS MidiScheduler class factory
         * @author  Pete Goodliffe
         * @version 1.00
         */
        class OSSMidiSchedulerFactory : public MidiSchedulerFactory
        {
            public:

                OSSMidiSchedulerFactory();
                virtual ~OSSMidiSchedulerFactory();

                /**
                 * @reimplemented
                 */
                virtual OSSMidiScheduler *createScheduler();

            private:
        };
    }
}

#endif
